import { compose, curry, equals, last, isArrayLike } from 'ramda'
import isPatch from './isPatch'
import listPath from './listPath'
const isLast = compose(equals('-'), last)

const applyCursorPatch = curry((cursor, patch) => {
  let updatePath = listPath(patch.path)
  let updateValue = patch.value
  let updateType

  let isLastEl = isLast(patch.path)

  if (isLastEl) {
    updatePath = dropLast(1, updatePath)
  }

  switch (patch.op) {
  case 'add':
    updateType = isLastEl ? 'push' : 'set'

    cursor.update(updatePath, {
      type: updateType,
      value: updateValue
    })

    break
  case 'replace':
    if (cursor.exists(updatePath)) {
      updateValue = isLastEl ? [-1, 1, updateValue] : updateValue
      updateType = isLastEl ? 'splice' : 'set'

      cursor.update(updatePath, {
        type: updateType,
        value: updateValue
      })

    } else {
      throw `Applying patch ${patch} failed. The path ${updatePath} does not exists.`
    }
    break
  case 'test':
    if (cursor.exists(updatePath)) {
      let currentValue = cursor.select(updatePath).get()
      if (!equals(currentValue, updateValue)) {
        throw `Applying patch ${patch} failed. The values ${currentValue} and ${updateValue} are not equal`
      }
    } else {
      throw `Applying patch ${patch} failed. The path ${updatePath} does not exists.`
    }
    break
  case 'move':
    console.error(`Applying patch ${patch} failed. Action "move" is not implemented`)
    break
  case 'copy':
    console.error(`Applying patch ${patch} failed. Action "copy" is not implemented`)
    break
  }

  return cursor
})


/**
 * Applies a JSON Patch to the given data cursor
 *
 * ** Sig: DataCursor -> Patch **
 *
 * @function
 * @implements curry
 * @param {object} dataCursor - The data cursor that will receive the patch
 * @param {object} patch - The JSON patch that implements https://tools.ietf.org/html/rfc6902
 * @return {object} cursor - The received cursor is returned after the applied patches
 * @example
 *
 * import { createPatch, applyBaobabJsPatch } from 'json-patch-utils'
 * import Baobab from 'baobab'
 *
 * let tree = new Baobab({
 *  foo: {
 *    bar: 123
 *  }
 * })
 * let patch = createPatch('add', '/foo/bar', 100)
 * applyPatch(tree, patch)
 * tree.get() // { foo: { bar: 100 } }
 *
 * @todo add atomic transactions
 */
const applyBaobabJsPatch = curry((cursor, xs) => {

  xs = isArrayLike(xs) ? xs : [xs]

  xs.forEach(x => {
    if (isPatch(x)) {
      applyCursorPatch(cursor, x)
    } else {
      throw 'Patch not valid'
    }
  })

  return cursor
})

export default applyBaobabJsPatch
